"""Algorithms for manipulating BibTeX data.

This module implements various algorithms supplied by BibTeX to style
files, as well as some algorithms to make BibTeX data more accessible
to Python.

Most of this file is originally part of biblib
(https://github.com/aclements/biblib)

Copyright (c) 2013 Austin Clements

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

__all__ = ('Name parse_names ' +
           'parse_month ' +
           'title_case ' +
           'TeXProcessor TeXToUnicode tex_to_unicode').split()

import re
import collections
import unicodedata
import pybtex.database as pybtex
import string
import stop_words

# Control sequences (defined as "control_seq_ilk" in bibtex) and their
# Unicode translations.  This is similar to, but slightly different
# from the TeX definitions (of course).
_CONTROL_SEQS = {
    '\\i': 'ı', '\\j': 'ȷ', '\\oe': 'œ', '\\OE': 'Œ',
    '\\ae': 'æ', '\\AE': 'Æ', '\\aa': 'å', '\\AA': 'Å',
    '\\o': 'ø', '\\O': 'Ø', '\\l': 'ł', '\\L': 'Ł', '\\ss': 'ß'
}

def single_entry_to_fulltext(entry: pybtex.Entry,
                             overwrite_key: str = None) -> str:
    """
    Converts a pybtex.Entry to text.

    :param entry: A pybtex.Entry.
    :param overwrite_key: A key to use instead of the default one.
    :return: A BibTeX entry as text.
    """
    effective_key = entry.key if not overwrite_key else overwrite_key
    formatter = pybtex.BibliographyData(entries={effective_key: entry})
    return formatter.to_string(bib_format="bibtex")

def fulltext_to_single_entry(fulltext: str) -> pybtex.Entry:
    """
    Parses a BibTeX entry into a pybtex.Entry
    """
    entry, = pybtex.parse_string(fulltext, bib_format="bibtex").entries.values()
    return entry

custom_key_skip_chars = str.maketrans("", "", " `~!@#$%^&*()+=[]{}|\\'\":;,<.>/?")
custom_key_skip_words = set(stop_words.get_stop_words("en"))
def generate_custom_key(entry: pybtex.Entry, key_format_in, suffix_level=0):
    key_format = key_format_in.replace('''{short_year}''', '''{short_year:02d}''')

    # TODO: fault tolerance against missing fields!
    year = int(entry.fields["year"])
    all_authors = parse_names(entry.fields["author"])
    author_surname = all_authors[0]\
        .pretty(template="{last}")\
        .lower()\
        .translate(custom_key_skip_chars)
    et_al = "_etAl" if len(all_authors) > 1 else ""

    filtered_title = [w for w in [t.lower() for t in entry.fields["title"].split()] if w not in custom_key_skip_words]
    if filtered_title:
        title_word = filtered_title[0]
    else:
        title_word = entry.fields["title"][0]
    title_word = title_word.translate(custom_key_skip_chars)

    return key_format.format(
        surname=author_surname,
        et_al=et_al,
        year=year,
        short_year=year%100,
        suffix='' if suffix_level==0 else chr(ord('a') + suffix_level - 1),
        title=title_word)

def field_to_unicode(entry, field, default_value=None):
    utf_field = entry.fields.get(field, default_value)
    try:
        utf_field = tex_to_unicode(utf_field)
    except:  # TODO: Limit the type of exceptions we catch
        pass
    return utf_field

class NameParser:
    def __init__(self):
        pass

    def __depth(self, data):
        depth, depths = 0, [0] * len(data)
        for pos, ch in enumerate(data):
            depths[pos] = depth
            if ch == '{':
                depth += 1
                depths[pos] = depth
            elif ch == '}':
                depth -= 1
        return depths

    def __split_depth0(self, regexp, data, flags=0):
        regexp = re.compile(regexp, flags=flags)
        depths = self.__depth(data)
        parts, last = [], 0
        for m in regexp.finditer(data):
            if depths[m.start()] == 0:
                parts.append(data[last:m.start()])
                last = m.end()
                if regexp.groups:
                    parts.extend(m.groups())
        parts.append(data[last:])
        return parts

    def _first_char(self, data):
        """Return the first character of data (in bibtex's sense)."""
        # XXX Should this be pulled out as some generic algorithm?
        pos = 0
        depths = self.__depth(data)
        while True:
            if pos == len(data):
                return ''
            elif data[pos].isalpha():
                return data[pos]
            elif data.startswith('{\\', pos):
                # Special character
                pos += 1
                m = re.compile(r'\\[a-zA-Z]+').match(data, pos)
                if m and m.group() in _CONTROL_SEQS:
                    # Known bibtex control sequence
                    return _CONTROL_SEQS[m.group()]
                # Scan for the first alphabetic character
                while pos < len(data) and depths[pos]:
                    if data[pos].isalpha():
                        return data[pos]
                    pos += 1
            elif data[pos] == '{':
                # Skip brace group
                while pos < len(data) and depths[pos]:
                    pos += 1
            else:
                pos += 1

    def __split_von_last(self, toks):
        # See von_name_ends_and_last_name_starts_stuff
        for von_end in range(len(toks) - 1, 1, -2):
            if self._first_char(toks[von_end - 2]).islower():
                return (toks[:von_end-1], toks[von_end:])
        return ([], toks)

    def parse(self, string):
        """Parse a BibTeX name list.

        Returns a list of Name objects.  Raises InputError if there is
        a syntax error.
        """

        # See x_format_name

        # Split names (see name_scan_for_and)
        name_strings = [n.strip() for n in self.__split_depth0(
            '[ \t]and(?=[ \t])', string, flags=re.IGNORECASE)]

        # Process each name
        names = []
        for name_string in name_strings:
            # Remove leading and trailing white space, ~, and -, and
            # trailing commas.
            name_string = name_trailing = name_string.lstrip('-~ \t')
            name_string = name_string.rstrip('-~ \t,')
            if ',' in name_trailing[len(name_string):]:
                # BibTeX warns about this because it often indicates a
                # bigger syntax problem
                # pos.warn('trailing comma after name `{}\''.format(name_string))
                pass

            # Split on depth-0 commas and further split tokens in each
            # part, keeping only the first connector between each
            # token.
            parts = [self.__split_depth0('([-~ \t])[-~ \t]*', part.strip())
                     for part in self.__split_depth0(',', name_string)]

            # Process name depending on how many commas there were
            first = von = last = jr = []
            if len(parts) == 1:
                # "First von Last"
                toks = parts[0]
                # The von tokens start with the first lower-case token
                # (but cannot start at the last token)
                for von_start in range(0, len(toks) - 2, 2):
                    if self._first_char(toks[von_start]).islower():
                        # Found beginning; now find the end
                        first = toks[:max(0, von_start-1)]
                        von, last = self.__split_von_last(toks[von_start:])
                        break
                else:
                    # No von tokens.  Find hyphen-connected last name
                    # tokens.
                    for last_start in range(len(toks) - 1, -1, -2):
                        if last_start and toks[last_start-1] != '-':
                            break
                    first = toks[:max(0, last_start-1)]
                    last = toks[last_start:]
            elif 2 <= len(parts) <= 3:
                # "von Last, First[, Jr]"
                von, last = self.__split_von_last(parts[0])
                first = parts[1]
                if len(parts) == 3:
                    jr = parts[2]
            else:
                raise ValueError('too many commas in name `{}\''.format(name_string))

            names.append(Name(''.join(first), ''.join(von),
                              ''.join(last), ''.join(jr)))
        return names

class Name(collections.namedtuple('Name', 'first von last jr')):
    """A parsed name.

    The name is parsed in to first name, "von", last name, and the
    complement (or "jr").  Each component is in uninterpreted form
    (e.g., TeX syntax).  Missing components are set to the empty
    string.
    """

    def is_others(self):
        return self.first == '' and self.von == '' and \
            self.last == 'others' and self.jr == ''

    def pretty(self, template='{first} {von} {last} {jr}'):
        """Pretty-print author according to template.

        The template is a 'format' template with the added feature
        that literal text surrounding fields that expand to empty
        strings is prioritized, rather than concatenated.
        Specifically, of the literal text snippets between two
        non-null fields, only the first of the highest priority is
        kept, where non-white space outranks white space outranks the
        empty string.  Literal text before and after the first and
        last fields is always kept.

        Hence, if the template is '{von} {last}, {first}, {jr}' and
        the name has a last and a jr not no von or first, then the
        first comma will be kept and the space and second dropped.  If
        the name has only a von and a last, then both commas will be
        dropped.  If the name has only a last, then all separators
        will be dropped.
        """

        # XXX BibTeX's own format.name$ templates are more
        # sophisticated than this, and it's not clear these are easier
        # to use.  These do have the (dubious) benefit of having
        # access to the usual format machinery.

        def priority(string):
            if not string:
                return 0
            elif string.isspace():
                return 1
            return 2
        fields = {'first': self.first, 'von': self.von,
                  'last': self.last, 'jr': self.jr}
        f = string.Formatter()
        pieces = ['']
        first_field, last_field = 0, -1
        leading = trailing = ''
        for i, (literal_text, field_name, format_spec, conv) in \
            enumerate(f.parse(template)):
            if i == 0:
                # Always keep leading text
                leading = literal_text
            elif field_name is None:
                # Always keep trailing test
                trailing = literal_text
            elif priority(literal_text) > priority(pieces[-1]):
                # Overrides previous piece
                pieces[-1] = literal_text

            if field_name is not None:
                obj, _ = f.get_field(field_name, (), fields)
                if not obj:
                    continue
                obj = f.convert_field(obj, conv)
                if first_field == 0:
                    first_field = len(pieces)
                last_field = len(pieces)
                pieces.extend([f.format_field(obj, format_spec), ''])
        # Only keep the pieces between non-null fields
        pieces = pieces[first_field:last_field + 1]
        return leading + ''.join(pieces) + trailing

def parse_names(string):
    """Parse a BibTeX name list (e.g., an author or editor field).

    Returns a list of Name objects.  The parsing is equivalent to
    BibTeX's built-in "format.name$" function.  Raises InputError if
    there is a syntax error.
    """
    return NameParser().parse(string)

_MONTHS = 'January February March April May June July August September October November December'.lower().split()

def parse_month(string):
    """Parse a BibTeX month field.

    This performs fairly fuzzy parsing that supports all standard
    month macro styles (and then some).

    Raises InputError if the field cannot be parsed.
    """
    val = string.strip().rstrip('.').lower()
    for i, name in enumerate(_MONTHS):
        if name.startswith(val) and len(val) >= 3:
            return i + 1
    raise ValueError('invalid month `{}\''.format(string))

CS_RE = re.compile(r'\\[a-zA-Z]+')

def title_case(string):
    """Convert to title case (like BibTeX's built-in "change.case$").

    Raises InputError if the title string contains syntax errors.
    """

    # See "@<Perform the case conversion@>"
    out = []
    level, prev_colon, pos = 0, False, 0
    while pos < len(string):
        keep = (pos == 0 or (prev_colon and string[pos-1] in ' \t\n'))

        if level == 0 and string.startswith('{\\', pos) and not keep:
            # Special character
            out.append(string[pos])
            pos += 1
            level += 1

            while level and pos < len(string):
                if string[pos] == '\\':
                    m = CS_RE.match(string, pos)
                    if m:
                        if m.group() in _CONTROL_SEQS:
                            # Lower case control sequence
                            out.append(m.group().lower())
                        else:
                            # Unknown control sequence, keep case
                            out.append(m.group())
                        pos = m.end()
                        continue
                elif string[pos] == '{':
                    level += 1
                elif string[pos] == '}':
                    level -= 1

                # Lower-case non-control sequence
                out.append(string[pos].lower())
                pos += 1

            prev_colon = False
            continue

        # Handle braces
        char = string[pos]
        if char == '{':
            level += 1
        elif char == '}':
            if level == 0:
                raise ValueError('unexpected }')
            level -= 1

        # Handle colon state
        if char == ':':
            prev_colon = True
        elif char not in ' \t\n':
            prev_colon = False

        # Change case of a regular character
        if level > 0 or keep:
            out.append(string[pos])
        else:
            out.append(string[pos].lower())
        pos += 1

    return ''.join(out)

# A TeX control sequence is
#
# 1) an active character (subsequent white space is NOT ignored) or,
# 2) a \ followed by either
# 2.1) a sequence of letter-category characters (subsequent white
#      space is ignored), or
# 2.2) a single space-category character (subsequent white space is
#      ignored), or
# 2.3) a single other character (subsequent white space is NOT
#      ignored).
#
# This regexp assumes plain TeX's initial category codes.  Technically
# only ~ and \f are active characters, but we include several other
# special characters that we want to abort on.
tex_cs_re = re.compile(
    r'([~\f$&#^_]|(\\[a-zA-Z]+|\\[ \t\r\n])|\\.)(?(2)[ \t\r\n]*)')

class TeXProcessor:
    """Base class for simple TeX macro processors.

    This assumes the initial category codes set up by plain.tex (and,
    likewise, LaTeX).
    """

    def process(self, string):
        """Expand active characters and macros in string.

        Raises InputError if it encounters an active character or
        macro it doesn't recognize.
        """

        self.__data = string
        self.__off = 0

        # Process macros
        while True:
            m = tex_cs_re.search(self.__data, self.__off)
            if not m:
                break
            self.__off = m.end()
            macro = m.group(1)
            nval = self._expand(macro)
            if nval is None:
                if macro.startswith('\\'):
                    raise ValueError('unknown macro `{}\''.format(macro))
                raise ValueError(
                    'unknown special character `{}\''.format(macro))
            self.__data = self.__data[:m.start()] + nval + \
                          self.__data[self.__off:]
            self.__off = m.start() + len(nval)

        return self.__data

    def _scan_argument(self):
        """Scan an return a macro argument."""
        if self.__off >= len(self.__data):
            raise ValueError('macro argument expected')
        if self.__data[self.__off] == '{':
            start = self.__off
            depth = 0
            while depth or self.__off == start:
                if self.__data[self.__off] == '{':
                    depth += 1
                elif self.__data[self.__off] == '}':
                    depth -= 1
                self.__off += 1
            return self.__data[start + 1:self.__off - 1]
        elif self.__data[self.__off] == '\\':
            m = tex_cs_re.match(self.__data, self.__off)
            self.__off = m.end()
            return m.group(1)
        else:
            arg = self.__data[self.__off]
            self.__off += 1
            return arg

    def _expand(self, cs):
        """Return the expansion of an active character or control sequence.

        Returns None if the sequence is unknown.  This should be
        overridden by sub-classes.
        """
        return None

class TeXToUnicode(TeXProcessor):
    """A simple TeX-to-unicode converter.

    This interprets accents and other special tokens like '--' and
    eliminates braces.
    """

    # Simple TeX-to-Unicode replacements
    _SIMPLE = {
        # Active characters
        '~': '\u00A0',
        # chardefs from plain.tex
        '\\%': '%', '\\&': '&', '\\#': '#', '\\$': '$', '\\ss': 'ß',
        '\\ae': 'æ', '\\oe': 'œ', '\\o': 'ø',
        '\\AE': 'Æ', '\\OE': 'Œ', '\\O': 'Ø',
        '\\i': 'ı', '\\j': 'ȷ',
        '\\aa': 'å', '\\AA': 'Å', '\\l': 'ł', '\\L': 'Ł',
        # Other defs from plain.tex
        '\\_': '_', '\\dag': '†', '\\ddag': '‡', '\\S': '§', '\\P': '¶',
    }

    # TeX accent control sequences to Unicode combining characters
    _ACCENTS = {
        # Accents defined in plain.tex
        '\\`': '\u0300', "\\'": '\u0301', '\\v': '\u030C', '\\u': '\u0306',
        '\\=': '\u0304', '\\^': '\u0302', '\\.': '\u0307', '\\H': '\u030B',
        '\\~': '\u0303', '\\"': '\u0308',
        '\\d': '\u0323', '\\b': '\u0331', '\\c': '\u0327',
        # Other accents that seem to be standard, but I can't find
        # their definitions
        '\\r': '\u030A', '\\k': '\u0328'
    }

    def process(self, string):
        string = super().process(string)

        # Handle ligatures that are unique to TeX.  This must be done
        # after macro expansion, but before brace removal because
        # braces inhibit ligatures.
        string = string.replace('---', '\u2014').replace('--', '\u2013')

        # Remove braces
        return string.replace('{', '').replace('}', '')

    def _expand(self, cs):
        if cs in self._SIMPLE:
            return self._SIMPLE[cs]
        if cs in self._ACCENTS:
            arg = self._scan_argument()
            if len(arg) == 0:
                seq, rest = ' ' + self._ACCENTS[cs], ''
            elif arg.startswith('\\i') or arg.startswith('\\j'):
                # Unicode combining marks should be applied to the
                # regular i, not the dotless i.
                seq, rest = arg[1] + self._ACCENTS[cs], arg[2:]
            else:
                seq, rest = arg[0] + self._ACCENTS[cs], arg[1:]
            return unicodedata.normalize('NFC', seq) + rest
        return None

def tex_to_unicode(string):
    """Convert a BibTeX field value written in TeX to Unicode.

    This interprets accents and other special tokens like '--' and
    eliminates braces.  Raises InputError if it encounters a macro it
    doesn't understand.

    Note that BibTeX's internal understanding of accented characters
    (e.g., purify$ and change.case$) is much more limited than TeX's.
    This implements something closer to TeX on the assumption that the
    goal is to display the string.
    """

    return TeXToUnicode().process(string)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Tool for downloading, maintaining, and searching a BibTeX database.

Authors:
- David Vilar <david.vilar@gmail.com>
- Matt Post <post@cs.jhu.edu>
"""

import argparse
import logging
import os
import re
import sys
from typing import List, Tuple
import urllib.request
import pybtex.database as pybtex
import subprocess
import tempfile
import textwrap
from tqdm import tqdm
import yaml

from pkg_resources import resource_filename, resource_exists

from .bibdb import BibDB
from . import bibutils
from .config import Config

VERSION = '0.3.8'

class BibsearchError(Exception):
    pass

try:
    # SIGPIPE is not available on Windows machines, throwing an exception.
    from signal import SIGPIPE

    # If SIGPIPE is available, change behaviour to default instead of ignore.
    from signal import signal, SIG_DFL
    signal(SIGPIPE, SIG_DFL)

except ImportError:
    logging.warning('Could not import signal.SIGPIPE (this is expected on Windows machines)')

BIBSETPREFIX="bib://"

OUTPUT_TYPES = ['txt', 'bib', 'md']
DEFAULT_OUTPUT_TYPE = 'txt'

def prompt(message: str, *answers_in: List[str], default=0, case_insensitive=True):
    valid_answers = [a.lower() for a in answers_in] if case_insensitive else answers_in
    single_letter_answers = [a[0] for a in valid_answers]
    assert len(valid_answers) == len(set(valid_answers)), "Answers are not unique"
    assert len(single_letter_answers) == len(set(single_letter_answers)), "Single letter answers are not unique"
    answer = input("%s [%s] " % (message, "/".join(answers_in)))
    answer_index = -1
    while answer_index < 0:
        if not answer and default >= 0:
            answer_index = default
        else:
            if case_insensitive:
                answer = answer.lower()
            try:
                answer_index = valid_answers.index(answer)
            except ValueError:
                try:
                    answer_index = single_letter_answers.index(answer)
                except ValueError:
                    answer = input("Please answer one of %s: " % "/".join(answers_in))
    return answers_in[answer_index]




def download_file(url, fname_out=None) -> None:
    """
    Downloads a file to a location.
    """

    import ssl

    #~ logging.info('Downloading {} to {}'.format(url, fname_out if fname_out is not None else 'STR'))
    # check if the file has already been downloaded
    if fname_out is not None and os.path.exists(fname_out):
        logging.info('Loading "%s" from cache at "%s"', url, fname_out)
        return fname_out
    elif fname_out:
        logging.info('Downloading "%s" -> "%s"', url, fname_out)
    else:
        logging.info('Downloading "%s"', url)

    try:
        with urllib.request.urlopen(url) as f:
            if not fname_out:
                return f.read().decode("utf-8")
            else:
                fdir = os.path.dirname(fname_out)
                if not os.path.exists(fdir):
                    os.makedirs(fdir)

                with open(fname_out, "wb") as outfile:
                    outfile.write(f.read())

                return fname_out

    except ssl.SSLError:
        logging.warning('An SSL error was encountered in downloading the files. If you\'re on a Mac, '
                        'you may need to run the "Install Certificates.command" file located in the '
                        '"Python 3" folder, often found under /Applications')
        sys.exit(1)


def format_search_results(results: List[Tuple[str,str]],
                          output_type=DEFAULT_OUTPUT_TYPE,
                          use_original_key=False) -> str:
    """
    Formats a set of (Pybtex.Entry, custom_key) pairs for printing to the terminal.
    Output can be either actual BibTeX or a summarized form.

    :param results: A list of pairs of full bibtex entries and their associated keys
    :param bibtex_output: Whether to retun the result as a bibtex entry instead of summarized text.
    :param use_original_key: Whether to use the default key instead of the custom one.
    :return: An amalgamated string including all entries.
    """

    output = ""
    for entryno, (fulltext, original_key) in enumerate(results, 1):
        entry = bibutils.fulltext_to_single_entry(fulltext)
        if use_original_key:
            entry.key = original_key
            fulltext = bibutils.single_entry_to_fulltext(entry)
        if output_type == 'bib':
            output += fulltext + "\n"
        else:
            utf_author = bibutils.field_to_unicode(entry, "author", "")
            utf_author = [a.pretty() for a in bibutils.parse_names(utf_author)]
            utf_author = ", ".join(utf_author[:-2] + [" and ".join(utf_author[-2:])])

            utf_title = bibutils.field_to_unicode(entry, "title", "")
            utf_venue = bibutils.field_to_unicode(entry, "journal", "")
            if not utf_venue:
                utf_venue = bibutils.field_to_unicode(entry, "booktitle", "")

            if output_type == 'txt':
                textwrapper = textwrap.TextWrapper(subsequent_indent="     ")
                lines = textwrapper.wrap('{index}. [{key}] {author} "{title}", {venue}{year}\n{url}'.format(
                                index=entryno,
                                key=entry.key,
                                author=utf_author,
                                title=utf_title,
                                venue=utf_venue + ", ",
                                year=entry.fields["year"],
                                url=entry.fields["url"]))
                output += "\n".join(lines) + "\n\n"

            elif output_type == 'md':
                output += '[{title}]({url})\n{authors}\n{venue}. {year}.'.format(
                    title=utf_title,
                    url=entry.fields["url"],
                    authors=utf_author,
                    venue=utf_venue,
                    year=entry.fields["year"]) + '\n\n'

    return output[:-1] # Remove the last empty line

def _get_cache_or_search_result(db: BibDB,
                                search_terms: List[str]) -> Tuple[str, str]:
    """
    If search_terms is empty or a single digit, return that item from
    the cache. Otherwise, return the first item from the search results.
    Item returned is a tuple (bibtex_str, custom_key).
    """
    if len(search_terms) == 0:
        # load the search cache, use first entry
        results = db.load_search_cache()
        if not results:
            logging.error("No documents found in search cache.")
            sys.exit(1)

        entry_index = 0

    elif len(search_terms) == 1 and re.match(r'\d+', search_terms[0]) is not None:
        # load the search cache, use specified entry
        results = db.load_search_cache()
        if not results:
            logging.error("No documents found in search cache.")
            sys.exit(1)

        entry_index = int(search_terms[0]) - 1

    else:
        results = db.search(args.terms)

        if not results:
            logging.error("Search returned no results.")
            sys.exit(1)

    if entry_index >= len(results):
        logging.error("You requested result {}, but only {} documents were found.".format(entry_index, len(results)))
        sys.exit(1)

    return results[entry_index]


def _find(args, config):
    db = BibDB(config)
    print(format_search_results(db.search(args.terms), args.output_type, args.original_key), end='')

def _open(args, config):
    """
    Opens the PDF associated with the last search query (if no
    arguments were given) or the requested index (if a single numeric
    argument was given) or the results of a search term (all other cases).

    """
    db = BibDB(config)
    bibtex, key = _get_cache_or_search_result(db, args.terms)

    entry = bibutils.fulltext_to_single_entry(bibtex)
    if "url" not in entry.fields:
        logging.error("Entry does not contain an URL field")
    if not os.path.exists(config.download_dir):
        os.makedirs(config.download_dir)
    temp_fname = download_file(entry.fields["url"], os.path.join(config.download_dir, entry.key + ".pdf"))
    subprocess.run([config.open_command, temp_fname])

def _download(args, config):
    db = BibDB(config)
    results = db.search(args.terms)
    if not results:
        logging.error("No documents returned by query")
        sys.exit(1)

    if not os.path.exists(config.download_dir):
        os.makedirs(config.download_dir)
    iterable = tqdm(results, ncols=80, bar_format="{l_bar}{bar}| [Elapsed: {elapsed} ETA: {remaining}]")
    for fulltext, _ in iterable:
        entry = bibutils.fulltext_to_single_entry(fulltext)
        if "url" not in entry.fields:
            logging.error("Entry does not contain an URL field")
        download_file(entry.fields["url"], os.path.join(config.download_dir, entry.key + ".pdf"))

class AddFileError(BibsearchError):
    pass

def _add_file(fname, force_redownload, db, per_file_progress_bar):
    """
    Return #added, #skipped, file_skipped
    """
    if fname.startswith('http'):
        if not force_redownload and db.file_has_been_downloaded(fname): 
            return 0, 0, True
        try:
            new_entries = pybtex.parse_string(download_file(fname),
                                              bib_format="bibtex").entries
        except urllib.error.URLError as e:
            raise AddFileError("Error downloading '%s' [%s]" % (fname, str(e)))
        except pybtex.PybtexError:
            raise AddFileError("Error parsing file %s" % fname)
        db.register_file_downloaded(fname)
    else:
        new_entries = pybtex.parse_file(fname,
                                        bib_format="bibtex").entries

    added = 0
    skipped = 0
    if per_file_progress_bar:
        iterable = tqdm(new_entries.values(), ncols=80, bar_format="{l_bar}{bar}| [Elapsed: {elapsed} ETA: {remaining}]")
    else:
        iterable = new_entries.values()
    for entry in iterable:
        if db.add(entry):
            added += 1
        else:
            skipped += 1

    return added, skipped, False

def get_fnames_from_bibset(raw_fname, database_url):
    bib_spec = raw_fname[len(BIBSETPREFIX):].strip()
    spec_fields = bib_spec.split('/')
    resource = spec_fields[0]
    if resource == "list":
        # Special case, list know resources
        textwrapper = textwrap.TextWrapper(subsequent_indent=10*" ")
        known_resources = download_file(database_url + "list.txt")
        for l in known_resources.split("\n"):
            line = l.strip()
            if line:
                name, description = line.split("\t")
                print("\n".join(textwrapper.wrap("%-10s%s" % (name, description))))
        return []
    uri = database_url + resource + '.yml'
    try:
        currentSet = yaml.load(download_file(uri))
        #~ currentSet = yaml.load(open("resources/" + resource + ".yml")) # for local testing
    except urllib.error.URLError as e:
        logging.error("Could not load resource %s from %s", resource, uri)
        logging.error('Maybe your connection is down?')
        sys.exit(1)
    if len(spec_fields) > 1:
        for f in spec_fields[1:]:
            # some keys are integers (years)
            try:
                currentSet = currentSet[f]
            except KeyError:
                logging.error("Invalid branch '%s' in bib specification '%s'",
                              f, raw_fname)
                logging.error("Options at this level are:", ', '.join(currentSet.keys()))
                sys.exit(1)
    def rec_extract_bib(dict_or_list):
        result = []
        if isinstance(dict_or_list, list):
            result = [fname for fname in dict_or_list]
        else:
            for v in dict_or_list.values():
                result += rec_extract_bib(v)
        return result
    return rec_extract_bib(currentSet)


def _arxiv(args, config):
    import feedparser

    db = BibDB(config)

    query = ' AND '.join(['"{}"'.format(x) for x in args.query])
    # move the query field search terms outside the quotes
    query = re.sub(r'"au(thor)?:', 'au:"', query)
    query = re.sub(r'"ti(tle)?:', 'ti:"', query)
    query = 'http://export.arxiv.org/api/query?{}'.format(urllib.parse.urlencode({ 'search_query': query,
                                                                                   'max_results': args.max_results }))
    logging.info('ARXIV QUERY: %s', query)
    response = download_file(query)

    feedparser._FeedParserMixin.namespaces['http://a9.com/-/spec/opensearch/1.1/'] = 'opensearch'
    feedparser._FeedParserMixin.namespaces['http://arxiv.org/schemas/atom'] = 'arxiv'
    feed = feedparser.parse(response)

    # print out feed information
    # print('Feed title: %s' % feed.feed.title)
    # print('Feed last updated: %s' % feed.feed.updated)

    # # print opensearch metadata
    # print('totalResults for this query: %s' % feed.feed.opensearch_totalresults)
    # print('itemsPerPage for this query: %s' % feed.feed.opensearch_itemsperpage)
    # print('startIndex for this query: %s'   % feed.feed.opensearch_startindex)

    # Run through each entry, and print out information
    results = []
    for entry in feed.entries:
        arxiv_id = re.sub(r'v\d+$', '', entry.id.split('/abs/')[-1])

        fields = { 'title': entry.title,
                   'journal': 'Computing Research Repository',
                   'year': str(entry.published[:4]),
                   'abstract': entry.summary,
                   'volume': 'abs/{}'.format(arxiv_id),
                   'archivePrefix': 'arXiv',
                   'eprint': arxiv_id,
        }

        try:
            fields['comment'] = entry.arxiv_comment
        except AttributeError:
            pass

        # get the links to the pdf
        for link in entry.links:
            try:
                if link.title == 'pdf':
                    fields['url'] = link.href
            except:
                pass

        authors = {'author': [pybtex.Person(author.name) for author in entry.authors]}
        bib_entry = pybtex.Entry('article', persons=authors, fields=fields)
        bib_entry.key = bibutils.generate_custom_key(bib_entry, config.custom_key_format)

        if args.add:
            db.add(bib_entry)
            results.append((bib_entry, bib_entry.key))
        else:
            results.append((bib_entry, arxiv_id))

        # Save the results to the search cache
        db.save_to_search_cache([(bibutils.single_entry_to_fulltext(bib_entry), key) for bib_entry, key in results])

    print(format_search_results( [(bibutils.single_entry_to_fulltext(bib_entry), arxiv_id)],
                                 bibtex_output=False,
                                 use_original_key=True))

    if args.add:
        db.save()


def _remove(args, config):
    db = BibDB(config)
    search_results = db.search(args.terms)
    if not search_results:
        logging.error("Search returned no results. Aborting.")
        sys.exit(1)
    print("You are about to delete these entries:")
    print("")
    print(format_search_results(search_results))
    confirmation = prompt("Do you want to proced with the deletion?", "yes", "NO",
                          default=1)
    if confirmation == "yes":
        for (_, original_key) in search_results:
            db.remove(original_key)
        db.save()
        print("Removed %d entries." % len(search_results))
    else:
        print("Aborted.")


def _add(args, config):
    db = BibDB(config)

    for raw_fname in args.files:
        fnames = [raw_fname] if not raw_fname.startswith(BIBSETPREFIX) \
                             else get_fnames_from_bibset(raw_fname, config.database_url)
        added = 0
        skipped = 0
        n_files_skipped = 0
        if len(fnames) > 1:
            iterable = tqdm(fnames, ncols=80, bar_format="Adding %s {l_bar}{bar}| [Elapsed: {elapsed} ETA: {remaining}]" % raw_fname)
            per_file_progress_bar = False
        else:
            iterable = fnames
            per_file_progress_bar = True
        error_msgs = []
        for f in iterable:
            try:
                f_added, f_skipped, file_skipped = _add_file(f, args.redownload, db, per_file_progress_bar)
                if args.verbose and not per_file_progress_bar:
                    if not file_skipped:
                        log_msg = "Added %d entries from %s" % (f_added, f)
                    else:
                        log_msg = "Skipped %s" % f
                    tqdm.write(log_msg)
            except AddFileError as e:
                f_added = 0
                f_skipped = 0
                file_skipped = False
                error_msgs.append(str(e))
            added += f_added
            skipped += f_skipped
            if file_skipped:
                n_files_skipped += 1

    print('Added', added, 'entries, skipped', skipped, 'duplicates. Skipped', n_files_skipped, 'files')
    if error_msgs:
        print("During operation followint errors occured:")
        for m in error_msgs:
            logging.error(m)
    db.save()

def _print(args, config):
    db = BibDB(config)
    if args.summary:
        print('Database has', len(db), 'entries')
    else:
        for entry in db:
            print(entry.rstrip() + "\n")

def _tex(args, config):
    citation_re = re.compile(r'\\citation{(.*)}')
    bibdata_re = re.compile(r'\\bibdata{(.*)}')
    db = BibDB(config)
    aux_fname = args.file
    if not aux_fname.endswith(".aux"):
        if aux_fname.endswith(".tex"):
            aux_fname = aux_fname[:-4] + ".aux"
        else:
            aux_fname = aux_fname + ".aux"
    bibfile = None
    entries = set()
    for l in open(aux_fname):
        match = citation_re.match(l)
        if match:
            keystr = match.group(1)
            for key in keystr.split(','):
                bib_entry = db.search_key(key)
                if bib_entry:
                    entries.add(bib_entry)
                else:
                    logging.warning("Entry '%s' not found", key)
        elif args.write_bibfile or args.overwrite_bibfile:
            match = bibdata_re.match(l)
            if match:
                bibfile = match.group(1)
    if bibfile:
        bibfile = os.path.join(os.path.dirname(aux_fname), bibfile+".bib")
        if os.path.exists(bibfile):
            if args.overwrite_bibfile:
                logging.info("Overwriting bib file %s.", bibfile)
            else:
                logging.error("Refusing to overwrite bib file %s. Use '-B' to force.", bibfile)
                sys.exit(1)
        else:
            logging.info("Writing bib file %s.", bibfile)
        fp_out = open(bibfile, "w")
    else:
        fp_out = sys.stdout
    for e in entries:
        print(e + "\n", file=fp_out)

def find_entry(entries_iter, field, value):
    for e in entries_iter:
        if e.fields.get(field) == value:
            return e
    return None

def compare_entries(old, new):
    added = set()
    deleted = set()
    edited = set()
    if old.key != new.key:
        edited.add("key")
    for old_field, old_value in old.fields.items():
        if not old_field in new.fields:
            deleted.add(old_field)
        else:
            if old_value != new.fields[old_field]:
                edited.add(old_field)
    added = set(new.fields.keys()) - set(old.fields.keys())
    return added, deleted, edited

def _edit(args, config):

    db = BibDB(config)
    bibtex, key = _get_cache_or_search_result(db, args.terms)

    with tempfile.NamedTemporaryFile("w") as temp_file:
        temp_fname = temp_file.name
        with open(temp_fname, "wt") as fp:
            original_entries_text = format_search_results(results=[(bibtex, key)],
                                                          output_type='bibtex',
                                                          use_original_key=False)
            fp.write(original_entries_text)
            original_entries = pybtex.parse_string(original_entries_text,
                                                   bib_format="bibtex").entries.values()
        logging.info('Using value of $EDITOR to choose editor.')
        subprocess.run([config.editor, temp_file.name])

        with open(temp_fname, "rt"):
            new_entries = pybtex.parse_file(temp_fname,
                                               bib_format="bibtex").entries.values()
    deleted_entries = []
    edited_entries = []
    seen_original_keys = set()
    changelog = []
    for new in new_entries:
        original_key = new.fields["original_key"]
        seen_original_keys.add(original_key)
        old = find_entry(original_entries,
                         "original_key",
                         original_key)
        added, deleted, edited = compare_entries(old, new)
        if added or deleted or edited:
            edited_entries.append(new)
            # Report changes
            edited_entries.append(new)
            changelog.append("\nEntry %s" % old.key)
            for field in added:
                changelog.append('\tAdded %s with value "%s"' % (field, new.fields[field]))
            for field in deleted:
                changelog.append("\tDeleted %s" % field)
            for field in edited:
                changelog.append('\tChanged %s to "%s"' %
                      (field,
                       new.key if field=="key" else new.fields[field]))
    for old in original_entries:
        if not old.fields["original_key"] in seen_original_keys:
            deleted_entries.append(old)
    if deleted_entries:
        changelog.append("\nDeleted entries:")
        for e in deleted_entries:
            changelog.append("\t%s" % e.key)

    if not edited_entries and not deleted_entries:
        logging.info("There were no changes in the entries.")
        sys.exit(0)

    print("Summary of changes:")
    print("\n".join(changelog) + "\n")

    confirmation = prompt("Do you want to perform these changes?", "YES", "no")
    if confirmation == "YES":
        for e in edited_entries:
            db.update(e)
        for e in deleted_entries:
            db.remove(e.key)
        db.save()
        print("Updated database.")
    else:
        print("Aborted.")

def _macros(args, config):
    for macro, expansion in config.macros.items():
        print("%s:\t%s" % (macro, expansion))

def _man(args, config):
    if not resource_exists('bibsearch', 'manual.1'):
        logging.error("Can't find manual page")
        sys.exit(1)

    manfile_path = resource_filename('bibsearch', 'manual.1')
    subprocess.run(["man", manfile_path])

# From https://stackoverflow.com/questions/13423540/argparse-subparser-hide-metavar-in-command-listing
class SubcommandHelpFormatter(argparse.RawDescriptionHelpFormatter):
    def _format_action(self, action):
        #parts = super(argparse.RawDescriptionHelpFormatter, self)._format_action(action)
        parts = super()._format_action(action)
        if action.nargs == argparse.PARSER:
            parts = "\n".join(parts.split("\n")[1:])
        return parts

def main():
    logging.basicConfig(level=logging.INFO,
                        format="[%(levelname)s] %(message)s")

    parser = argparse.ArgumentParser(description="bibsearch: Download, manage, and search a BibTeX database.\nUse '%(prog)s man' to get complete help.",
                                     formatter_class=SubcommandHelpFormatter)
    parser.add_argument('--version', '-V', action='version', version='%(prog)s {}'.format(VERSION))
    parser.add_argument('-c', '--config_file', help="use this config file",
                        default=os.path.join(os.path.expanduser("~"),
                                             '.bibsearch',
                                             "config")
                        )
    parser.set_defaults(func=lambda *_ : parser.print_help())
    subparsers = parser.add_subparsers(title="commands",
                                       description="Use '%(prog)s <command> -h' to obtain additional help.",
                                       metavar="<command>")

    parser_add = subparsers.add_parser('add', help='Add a BibTeX file')
    parser_add.add_argument('files', type=str, default=None, help='BibTeX files to add', nargs='+')
    parser_add.add_argument("-r", "--redownload", help="Re-download already downloaded files", action="store_true")
    parser_add.add_argument("-v", "--verbose", help="Be verbose about which files are being downloaded", action="store_true")
    parser_add.set_defaults(func=_add)

    parser_arxiv = subparsers.add_parser('arxiv', help='Search the arXiv')
    parser_arxiv.add_argument('query', type=str, nargs='+', default=None, help='Search query')
    parser_arxiv.add_argument("-m", "--max-results", type=int, default=10, help="Maximum number of results to return") 
    parser_arxiv.add_argument("-a", "--add", action='store_true', help="Add all results to the database (default: just print them to STDOUT)")
    parser_arxiv.add_argument("--output-type", "-o", default=DEFAULT_OUTPUT_TYPE, choices=OUTPUT_TYPES, help="Output type. Default: %(default)s")
    parser_arxiv.set_defaults(func=_arxiv)

    parser_dump = subparsers.add_parser('print', help='Print the BibTeX database')
    parser_dump.add_argument('--summary', action='store_true', help='Just print a summary')
    parser_dump.set_defaults(func=_print)

    parser_find = subparsers.add_parser('find', help='Search the database using fuzzy syntax', aliases=['search'])
    parser_find.add_argument('-b', '--bibtex', help='Print entries in bibtex format', action='store_true')
    parser_find.add_argument('--original-key', help='Print the original key of the entries', action='store_true')
    parser_find.add_argument('--output-type', '-o', default=DEFAULT_OUTPUT_TYPE, choices=OUTPUT_TYPES, help="Output type. Default: %(default)s")
    parser_find.add_argument('terms', nargs='*', help="One or more search terms which are ANDed together")
    parser_find.set_defaults(func=_find)

    parser_open = subparsers.add_parser('open', help='Open the article, if search returns only one result and url is available')
    parser_open.add_argument('terms', nargs='*', help="One or more search terms which are ANDed together")
    parser_open.set_defaults(func=_open)

    parser_download = subparsers.add_parser('download', help='download articles')
    parser_download.add_argument('terms', nargs='*', help="One or more search terms which are ANDed together")
    parser_download.set_defaults(func=_download)

    parser_tex = subparsers.add_parser('tex', help='Create .bib file for a latex article')
    parser_tex.add_argument('file', help='Article file name or .aux file')
    parser_tex.add_argument('-b', '--write-bibfile', help='Autodetect and write bibfile', action='store_true')
    parser_tex.add_argument('-B', '--overwrite-bibfile', help='Autodetect and write bibfile', action='store_true')
    parser_tex.set_defaults(func=_tex)

    parser_edit = subparsers.add_parser('edit', help='Edit entries')
    parser_edit.add_argument('terms', nargs='*', help='One or more search terms')
    parser_edit.set_defaults(func=_edit)

    parser_rm = subparsers.add_parser('remove', help='Remove an entry', aliases=['rm'])
    parser_rm.add_argument('terms', nargs='*', help='One or more search terms')
    parser_rm.set_defaults(func=_remove)

    parser_macros = subparsers.add_parser('macros', help='Show defined macros')
    parser_macros.set_defaults(func=_macros)

    parser_man = subparsers.add_parser('man', help='Shows documentation in form of a man page', aliases=['help'])
    parser_man.set_defaults(func=_man)

    args = parser.parse_args()
    config = Config(args.config_file)
    args.func(args, config)

if __name__ == '__main__':
    main()



import configparser
import logging
import os
import platform
import tempfile

class Config():
    defaults = {
        "bibsearch" : {
              "bibsearch_dir": os.path.expanduser("~/.bibsearch")
            , "download_dir": os.path.expanduser("~/.bibsearch/papers")
            , "open_command": "xdg-open" if platform.system() == "Linux" else "open"
            , "database_url": "https://github.com/mjpost/bibsearch/raw/master/resources/"
            , "custom_key_format": "{surname}{year}{suffix}:{title}"
            , "editor": os.environ.get("EDITOR", "nano")
        }
        , "macros" : {
              '@acl': 'venue:"Annual Meeting of the Association for Computational Linguistics"'
            , '@emnlp': 'venue:"Conference on Empirical Methods in Natural Language Processing"'
            , '@wmt': '(venue:"Workshop on Statistical Machine Translation" OR venue:"Conference on Machine Translation")'
            , '@naacl': 'venue:"Conference of the North American Chapter of the Association for Computational Linguistics"'
            , '@cl': 'venue:"Computational Linguistics"'  # Probably useless because the terms comes up in several conferences as well
            , '@arxiv': 'venue:"Computing Research Repository"'
            , '@corr': 'venue:"Computing Research Repository"'
        }
    }

    def __init__(self, fname):
        config = configparser.ConfigParser()
        config.read_dict(self.__class__.defaults)
        config.read(fname)
        for k, v in config["bibsearch"].items():
            if k in self.__class__.defaults["bibsearch"]:
                self.__setattr__(k, v)
            else:
                logging.warning("Unknown config option '%s'", k)

        self.macros = config["macros"]



import functools
import logging
import os.path
import pybtex.database as pybtex
import sqlite3
import sys
import yaml

from typing import Tuple

from . import bibutils

class BibDB:
    def __init__(self, config):
        self.config = config
        self.fname = os.path.join(self.config.bibsearch_dir, "bib.db")

        self.column_names_no_key = ["author", "title", "venue", "year"]

        createDB = False
        if not os.path.exists(self.fname):
            if not os.path.exists(os.path.dirname(self.fname)):
                os.makedirs(os.path.dirname(self.fname))
            createDB = True
        self.connection = sqlite3.connect(self.fname)
        self.cursor = self.connection.cursor()
        if createDB:
            self._create_db()
        # Find out if we have FTS
        self.cursor.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='bibindex'")
        self.has_fts = bool(self.cursor.fetchone())

    def _create_db(self):
        self.cursor.execute("""CREATE TABLE bib (
            key text UNIQUE,
            custom_key text UNIQUE,
            author text,
            title text,
            venue text,
            year text,
            fulltext text
            )""")
        self.cursor.execute("""CREATE TABLE downloaded_files (
            file text UNIQUE
            )""")
        try:
            self.cursor.executescript(
            """CREATE VIRTUAL TABLE bibindex USING fts5(
                key,
                custom_key,
                author,
                title,
                venue,
                year,
                fulltext UNINDEXED,
                content='bib',
                );
            CREATE TRIGGER bib_ai AFTER INSERT ON bib BEGIN
               INSERT INTO bibindex
                   (rowid, key, custom_key, author, title, venue, year, fulltext)
                   VALUES 
                   (new.rowid, new.key, new.custom_key, new.author, new.title, 
                   new.venue, new.year, new.fulltext);
                END;
            CREATE TRIGGER bib_ad AFTER DELETE ON bib BEGIN
               INSERT INTO bibindex
                   (bibindex, rowid, key, custom_key, author, title, venue, year, fulltext)
                   VALUES 
                   ('delete', old.rowid, old.key, old.custom_key, old.author, old.title, 
                   old.venue, old.year, old.fulltext);
                END;
            CREATE TRIGGER bibindex_au AFTER UPDATE ON bib BEGIN
               INSERT INTO bibindex
                   (bibindex, rowid, key, custom_key, author, title, venue, year, fulltext)
                   VALUES 
                   ('delete', old.rowid, old.key, old.custom_key, old.author, old.title, 
                   old.venue, old.year, old.fulltext);
               INSERT INTO bibindex
                   (rowid, key, custom_key, author, title, venue, year, fulltext)
                   VALUES 
                   (new.rowid, new.key, new.custom_key, new.author, new.title, 
                   new.venue, new.year, new.fulltext);
                END;
            """)
        except sqlite3.OperationalError as e:
            error_msg = str(e)
            if "no such module" in error_msg and "fts5" in error_msg:
                logging.warning("It seems your sqlite3 installation does not support fts5 indexing.")
                logging.warning("It is strongly encouraged to activate fts5 (see the README and the FAQ).")
                input("Press ENTER to continue...")
            else:
                raise

    def __len__(self):
        self.cursor.execute('SELECT COUNT(*) FROM bib')
        return int(self.cursor.fetchone()[0])

    def save_to_search_cache(self, search: Tuple[str,str,str]) -> None:
        """
        Saves the result of a search, which is represented as a tuple: (full bibtex, collection, key).
        This can then be used by subcommands that support implicit arguments, like 'open'.
        """
        last_results_fname = os.path.join(self.config.bibsearch_dir, "lastSearch.yml")
        with open(last_results_fname, "w") as fp:
            yaml.dump(search, fp)

    def load_search_cache(self) -> Tuple[str,str,str]:
        """
        Returns the result of a cached search.
        """
        last_results_fname = os.path.join(self.config.bibsearch_dir, "lastSearch.yml")
        if os.path.exists(last_results_fname):
            return yaml.load(open(last_results_fname))

    def _format_query_fts(self, query_terms):
        processed_query_terms = []
        for t in query_terms:
            current_term = t
            if t in self.config.macros:
                current_term = self.config.macros[t]
                # TODO: for now we trust macros blindly
            else:
                if not (current_term.startswith("author:") or
                        current_term.startswith("key:") or
                        current_term.startswith("title:") or
                        current_term.startswith("venue:") or
                        current_term.startswith("year")):
                    # Protect the whole sequence
                    if current_term[0] != '"' and current_term[-1] != '"':
                        current_term = '"%s"' % current_term
                else:
                    specifier, query = current_term.split(":", 1)
                    quoted_query = query if (query[0] == '"' and query[-1] == '"') \
                                         else '"%s"' % query
                    if specifier == "key":
                        current_term = '(key:%s OR custom_key:%s)' % (quoted_query, quoted_query)
                    else:
                        current_term = '%s:%s' % (specifier, quoted_query)
            processed_query_terms.append(current_term)
        return " AND ".join(processed_query_terms)

    def _format_query_no_fts(self, input_terms):
        query_terms = []
        query_values = []
        for t in input_terms:
            if not functools.reduce(lambda x, y: x or y,
                                    [t.startswith(c + ":")
                                     for c in self.column_names_no_key + ["key"]]):
                current_terms = []
                for c in self.column_names_no_key + ["key", "custom_key"]:
                    current_terms.append('(%s LIKE ?)' % c)
                    query_values.append('%%%s%%' % t)
                query_terms.append("(%s)" % " OR ".join(current_terms))
            else:
                specifier, query = t.split(":", 1)
                wildquery = '%%%s%%' % query
                if specifier == "key":
                    query_terms.append('((key LIKE ?) OR (custom_key LIKE ?))')
                    query_values.append(wildquery)
                    query_values.append(wildquery)
                else:
                    query_terms.append('(%s LIKE ?)' % specifier)
                    query_values.append(wildquery)
        return " AND ".join(query_terms), query_values

    def search(self, query: str):
        """
        Performs a search against the private database.

        :param query: The search query.
        :return: A list of search results.
        """
        if self.has_fts:
            self.cursor.execute("SELECT fulltext, key FROM bibindex \
                                WHERE bibindex MATCH ?",
                                [self._format_query_fts(query)])
        else:
            where_clause, query_values = self._format_query_no_fts(query)
            self.cursor.execute("SELECT fulltext, key FROM bib \
                                    WHERE %s" % where_clause,
                                query_values)
        results = list(self.cursor)

        self.save_to_search_cache(results)

        return results

    def search_key(self, key) -> str:
        """
        Searches the database on the specified key or custom key.
        Returns the fulltext entry with the queried key as the entry key.

        :param key: The key to search on (key or custom key)
        :return: The full-text entry.
        """
        self.cursor.execute("SELECT fulltext FROM bib WHERE key=? OR custom_key=?",
                            [key, key])
        entry = self.cursor.fetchone()
        if entry is not None:
            entry = bibutils.single_entry_to_fulltext(bibutils.fulltext_to_single_entry(entry[0]), overwrite_key=key)
        return entry

    def save(self):
        self.connection.commit()

    def remove(self, key: str):
        """Removes the entry"""

        self.cursor.execute('DELETE FROM bib WHERE key=? or custom_key=?', [key, key])

    def add(self, entry: pybtex.Entry):
        """ Returns if the entry was added or if it was a duplicate"""

        # TODO: make this a better sanity checking and perhaps report errors
        if not entry.key:
            return False
        if not entry.fields.get("author"):
            entry.fields["author"] = "UNKNOWN"

        original_key = entry.key
        entry.fields["original_key"] = original_key
        utf_author = bibutils.field_to_unicode(entry, "author")
        utf_title = bibutils.field_to_unicode(entry, "title")
        utf_venue = bibutils.field_to_unicode(entry, "journal")
        if not utf_venue:
            utf_venue = bibutils.field_to_unicode(entry, "booktitle")
        custom_key_tries = 0
        added = False
        while not added:
            custom_key = None
            if custom_key_tries < 27:
                try:
                    custom_key = bibutils.generate_custom_key(entry, self.config.custom_key_format, custom_key_tries)
                except Exception as e:
                    pass
            else:
                logging.warning("Could not generate a unique custom key for entry %s", original_key)
                custom_key = original_key
            try:
                self.cursor.execute('INSERT INTO bib(key, custom_key, author, title, venue, year, fulltext) VALUES (?,?,?,?,?,?,?)',
                                    (original_key,
                                     custom_key,
                                     utf_author,
                                     utf_title,
                                     utf_venue,
                                     str(entry.fields.get("year")),
                                     bibutils.single_entry_to_fulltext(entry, custom_key)
                                    )
                                   )
                added = True
            except sqlite3.IntegrityError as e:
                error_message = str(e)
                if "UNIQUE" in error_message:
                    if "bib.custom_key" in error_message:
                        # custom_key was already in the DB
                        custom_key_tries += 1
                    elif "bib.key" in error_message:
                        # duplicate entry
                        break
                    else:
                        raise
                else:
                    raise
        return added

    def update_custom_key(self, original_key, new_custom_key):
        self.cursor.execute("SELECT fulltext FROM bib WHERE key=? LIMIT 1", (original_key,))
        entry = bibutils.fulltext_to_single_entry(self.cursor.fetchone()[0])
        entry.key = new_custom_key
        try:
            self.cursor.execute("UPDATE bib SET custom_key=?, fulltext=? WHERE key=?",
                                [new_custom_key,
                                 bibutils.single_entry_to_fulltext(entry),
                                 original_key])
            self.save()
        except:
            logging.error("Key %s already exists in the database", new_custom_key)
            sys.exit(1)

    def update(self, entry: pybtex.Entry):
        """ Returns if the entry was added or if it was a duplicate"""

        # TODO: make this a better sanity checking and perhaps report errors
        if not entry.key:
            return False

        original_key = entry.fields["original_key"]
        utf_author = bibutils.field_to_unicode(entry, "author")
        utf_title = bibutils.field_to_unicode(entry, "title")
        utf_venue = bibutils.field_to_unicode(entry, "journal")
        if not utf_venue:
            utf_venue = bibutils.field_to_unicode(entry, "booktitle")
        try:
            self.cursor.execute('UPDATE bib SET custom_key=?, author=?, title=?, venue=?, year=?, fulltext=? WHERE key=?',
                                (entry.key,
                                 utf_author,
                                 utf_title,
                                 utf_venue,
                                 str(entry.fields.get("year")),
                                 bibutils.single_entry_to_fulltext(entry),
                                 original_key
                                )
                               )
        except sqlite3.IntegrityError as e:
            error_message = str(e)
            if "UNIQUE" in error_message:
                if "bib.custom_key" in error_message:
                    logging.error("Key %s already exists in the database")
                else:
                    raise
            else:
                raise

    def __iter__(self):
        self.cursor.execute("SELECT fulltext FROM bib")
        for e in self.cursor:
            yield e[0]

    def file_has_been_downloaded(self, file):
        return self.cursor.execute("""SELECT 1 FROM downloaded_files WHERE file = ? LIMIT 1""", [file]).fetchone() is not None

    def register_file_downloaded(self, file):
        try:
            self.cursor.execute("""INSERT INTO downloaded_files(file) VALUES (?)""", [file])
        except sqlite3.IntegrityError:
            # File was already registered. No problem.
            pass



#!/usr/bin/env python3

# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
A setuptools based setup module.

    python3 setup.py sdist upload -r pypi

See:
- https://packaging.python.org/en/latest/distributing.html
- https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import re
import os
import sys
import subprocess

ROOT = os.path.dirname(__file__)

def get_long_description():
    with open(os.path.join(ROOT, 'README.md'), encoding='utf-8') as f:
        return f.read()

def get_version():
    VERSION_RE = re.compile(r'''VERSION\s+=\s+['"]([0-9.]+)['"]''')
    init = open(os.path.join(ROOT, 'bibsearch', 'bibsearch.py')).read()
    return VERSION_RE.search(init).group(1)

setup(
    name = 'bibsearch',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version = get_version(),

    description = 'Download, manage, and search a BibTeX database.',
    long_description = get_long_description(),

    # The project's main homepage.
    url = 'https://github.com/mjpost/bibsearch',

    author = 'David Vilar, Matt Post',
    author_email='david.vilar@gmail.com, post@cs.jhu.edu',
    maintainer_email='david.vilar@gmail.com, post@cs.jhu.edu',

    license = 'Apache License 2.0',

    python_requires = '>=3',

    # Alternatively, if you want to distribute just a my_module.py, uncomment this:
    # py_modules = ["bibsearch", "bibdb", "bibutils", "config"],
    packages = find_packages(),

    package_data = { 'bibsearch': ['manual.1'] },

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires = ['typing', 'tqdm', 'pyaml', 'stop-words', 'pybtex', 'feedparser'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require = {},

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'bibsearch = bibsearch.bibsearch:main',
        ],
    },

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers = [
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Text Processing',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3 :: Only',
    ],

    # What does your project relate to?
    keywords = ['computer science, LaTeX, BibTeX'],
)
